from .protectors.DatabaseProtector import DatabaseProtector

from .protectors.IPProtector import IPProtector
from .protectors.BotProtector import BotProtector


class Security:
    def __init__(self, application) -> None:
        self.app = application

    def secure(self, request):
        try:
            (
                # bot protection
                BotProtector(self.app, request)
                .block_fake_bot()
                .block_bot()
                .block()
            )

            (
                # ip protection
                IPProtector(self.app, request)
                .block_ip()
                .throttle()
            )

            (
                # database protection
                DatabaseProtector(self.app, request)
            )
        except Exception as e:
            return self.app.make("response").status(403).json({"message": str(e)})

        return request



"""A SecurityProvider Service Provider."""

from masonite.packages import PackageProvider
from ..Security import Security
from ..middlewares.SecurityMiddleware import SecurityMiddleware


class SecurityProvider(PackageProvider):
    def configure(self):
        """Register objects into the Service Container."""
        (self.root("security").name("security").config("config/security.py", publish=True))

    def register(self):
        super().register()
        self.application.bind("security", Security(application=self.application))
        self.application.make("middleware").add({"protect": [SecurityMiddleware]})

    def boot(self):
        """Boots services required by the container."""
        pass



"""Security Settings"""

"""
|--------------------------------------------------------------------------
| Masonite Security
|--------------------------------------------------------------------------
|
"""

BLOCK_IP = True
BLOCK_BOT = True
THROTTLE_REQUESTS = True

MAX_REQUESTS = 20  # requests per IP address per minute (default: 20)
IP_BLOCK_DURATION = 60  # seconds (default: 60)

# list of IP addresses to block (default: [])
BLOCKED_IPS = []

# list of Bot Agents to block
BLOCKED_BOTS = [
    "AhrefsBot",
    "Alexibot",
    "Baiduspider",
    "BlackWidow",
    "BLEXBot",
    "dotbot",
    "Jetbot",
    "MJ12bot",
    "Nutch",
    "rogerbot",
    "SemrushBot",
    "SemrushBot-SA",
    "Standford",
    "SurveyBot",
    "WebVac",
    "serpstatbot",
]



from masonite.middleware import Middleware


class SecurityMiddleware(Middleware):
    def before(self, request, response):
        """Return the request."""
        print("i am called...")
        return request.app.make("security").secure(request)

    def after(self, request, response):
        """Return the response."""
        return response



from masonite.configuration import config


class BaseProtector:
    def __init__(self, application, request) -> None:
        self.application = application
        self.request = request
        self.security_config = config("security")
        self.user_agent = request.environ.get("HTTP_USER_AGENT", "").lower()
        self.blocked = False
        self.cache = application.make("cache").store()

    def ip(self):
        return self.request.ip() or "127.0.0.1"

    def block(self):
        pass



from ipaddress import ip_address
from .BaseProtector import BaseProtector
import re
import socket

BOTS = (
    # Bot signature string, bot verification regex
    ("baiduspider", re.compile(r".*crawl\.baidu\.(com|jp)$")),
    ("googlebot", re.compile(r".*(google|googlebot)\.com$")),
    ("bingbot", re.compile(r".*search\.msn\.com$")),
    ("slurp", re.compile(r".*crawl\.yahoo\.net$")),
    ("yandexbot", re.compile(r".*yandex\.(ru|com|net)$")),
    ("msnbot", re.compile(r".*msn\.com$")),
    ("duckduckbot", re.compile(r".*duckduckgo\.com$")),
    ("sogou", re.compile(r".*sogou\.com$")),
)


class BotProtector(BaseProtector):
    def __init__(self, application, request):
        super().__init__(application, request)

    def block_fake_bot(self):
        if self.security_config.get("block_bot", True) is False:
            return self

        matched_signatures = [
            signature for signature in BOTS if signature[0].lower() in self.user_agent
        ]
        if len(matched_signatures) == 1:
            bot_sig_string, bot_verification_regex = matched_signatures[0]
            client_ip = self.request.ip()

            if not client_ip:
                raise Exception("IP spoofing detected.")

            if not self.ipaddress_is_private(ip=client_ip):
                try:
                    bot_reverse_lookup = self.reverse_lookup(client_ip)
                except socket.herror as e:
                    if e.errno == 1:
                        raise Exception("IP spoofing detected.")
                else:
                    is_fake_bot = bot_verification_regex.match(bot_reverse_lookup) is None
                    if is_fake_bot:
                        raise Exception("IP spoofing detected.")
        return self

    def block_bot(self):
        if self.security_config.get("block_bot", True) is False:
            return self

        bots = self.security_config.get("blocked_bots", [])
        for bot in bots:
            if bot in self.user_agent:
                raise Exception("SPAM bot detected.")
        return self

    def ipaddress_is_private(self, ip):
        return ip_address(address=ip).is_private

    def reverse_lookup(self, ip_address):
        return socket.gethostbyaddr(ip_address)[0]



import pendulum
from .BaseProtector import BaseProtector


class IPProtector(BaseProtector):
    def key(self):
        return f"ip_blocker_{self.ip()}"

    def ip_info(self):
        if self.cache.has(self.key()):
            old = self.cache.get(self.key())
            self.cache.put(
                self.key(),
                {
                    "ip": self.ip(),
                    "blocked": old.get("blocked", False),
                    "counter": old.get("counter", 0) + 1,
                    "started_at": old.get("started_at", pendulum.now().timestamp()),
                },
            )
            return old
        else:
            data = {
                "ip": self.ip(),
                "counter": 1,
                "blocked": False,
                "started_at": pendulum.now().int_timestamp,
            }
            self.cache.put(self.key(), data)
            return data

    def block_ip(self):
        """Block IP address."""
        if self.security_config.get("block_ip", True) is False:
            return self

        ips = self.security_config.get("blocked_ips", [])
        if len(ips) > 0:
            if self.ip() in ips:
                raise Exception("IP address is blocked.")
        return self

    def throttle(self):
        """Throttle IP address."""
        if self.security_config.get("throttle_requests", True) is False:
            return self

        info = self.ip_info()
        gap = pendulum.now().int_timestamp - info["started_at"]
        block_duration = self.security_config.get("ip_block_duration", 60)

        if info["blocked"] is True:
            if gap <= block_duration:
                raise Exception("IP address is blocked.")
            else:
                self.cache.forget(self.key())
                return self

        if gap < block_duration:
            if info["counter"] > self.security_config.get("max_requests", 10):
                info["blocked"] = True
                self.cache.put(self.key(), info)
        else:
            self.cache.forget(self.key())

        return self

    def block(self):
        pass
